#ifndef MICROARRAY_HPP_
#define MICROARRAY_HPP_

#include"config.h"
#include"constants.h"
#include"genes.hpp"

#include<string>
#include<list>
#include<map>
#include<vector>
#include<stdlib.h>
#include<string.h>
#include<math.h>

namespace allinone {

  using namespace std;

  class Microarray;

  class Gene{
    public:
   
    int v;

    Microarray * m;

    __INLINE bool operator!=(const Gene & g2) const {
        return m != g2.m || v != g2.v;
    }

    __INLINE bool operator==(const Gene & g2) const {
        return m == g2.m && v == g2.v ;
    }

    public:
    Gene (){}
    __INLINE Gene (Microarray * m, int _1){
      v = _1;
      this->m = m;
    }

    __INLINE _PFloat & operator[] (int vector_no);

    const char * get_name();
  };

  template<int reverse>
  class Gene_iterator :  public 
      allinone::Iterator<Gene, reverse, Gene_iterator<reverse> >{
  public:
    __INLINE Gene_iterator (const Gene & g) : 
        allinone::Iterator<Gene, reverse, Gene_iterator<reverse> >(g) {}

    __INLINE Gene_iterator<reverse> & operator+= (const int & _n){
        (*this)->v += _n * (reverse ? -1 : 1);
        return *this;
    }
  };


  class Microarray : public Genes{
    public:
    typedef Gene_iterator<0> iterator;
    typedef Gene_iterator< 1> reverse_iterator;
    
    __INLINE int get_vector_length(){return vector_length;}
    __INLINE int get_no_genes()  {return no_genes;  }

    private:
    const int no_genes;
    const int vector_length; 
    
     
    vector<const char *> experiments_names;
    map<const char *, int, allinone::charptr_less> experiments_names_to_i;

    _PFloat * microarray_array;

    void init (){
      long dataSize = (long)this->no_genes * this->vector_length * sizeof(_PFloat);
      microarray_array = (_PFloat *)malloc(dataSize);
      memset(microarray_array, 0, dataSize);
      experiments_names.resize(this->vector_length);
    }

    public:

    Microarray(int vector_length, int no_genes) : 
    Genes(no_genes, InterfaceFlow::Microarray),
    no_genes(no_genes), vector_length(vector_length){
        init();
    }

    Microarray(const Microarray * m) : 
    Genes(m->no_genes, m),
    no_genes(m->no_genes), vector_length(m->vector_length){
        init();
        Microarray * m2 = (Microarray *) m;
        this->set_names(&m2->get_name(0), no_genes, 0);
        this->set_descs(&m2->get_desc(0), no_genes, 0);
        memcpy(&(*this)[0][0], &(*m2)[0][0], vector_length * no_genes * sizeof(_PFloat));
    }

    virtual long get_no_vertices() const{
      return no_genes;
    }
        
    virtual ~Microarray(){

    }
    
    __INLINE _PFloat & get(int gene_no, int vector_no){
#ifdef __DEBUG__
      if(gene_no >= no_genes || vector_no >= vector_length)
        LOG(WARNING, "Trying to access gene: %d:%d which is not found in MA", 
            gene_no, vector_no);
#endif 
      return microarray_array[gene_no * vector_length + vector_no];
    }

    __INLINE _PFloat & get(const char * gene, int vector_no){
#ifdef __DEBUG__
      if(vertices_names_to_i.find(gene) == vertices_names_to_i.end())
        LOG(WARNING, "Trying to access gene: %s which is not found in MA", gene);
#endif 
      return get(this->get_no(gene), vector_no);
    }

    __INLINE _PFloat & get(const char * gene, const char * vector){
#ifdef __DEBUG__
      if(experiments_names_to_i.find(vector) == experiments_names_to_i.end())
        LOG(WARNING, "Trying to access experiment: %s which is not found in MA",
           vector);
#endif 
      return get(gene, this->get_exp_no(vector));
    }


    void set_exps(const char ** names, int no, int offset = 0){
      const char * tmp;
      map<const char *, int, allinone::charptr_less>::iterator it;
      for(int i = offset; i < offset + no; i++){
        if((tmp = experiments_names[i]) &&
           (it = experiments_names_to_i.find(tmp)) != experiments_names_to_i.end()){
           experiments_names_to_i.erase(it);
           free((void *)tmp);
        }
        experiments_names[i] = tmp = strdup(names[i - offset]);
        experiments_names_to_i[tmp] = i;
      }
    }

    const char * & get_exp(int exp) {
      return experiments_names[exp];
    }

    int get_exp_no(const char * name) {
      return experiments_names_to_i.find(name)->second;
    }


     void normalize(){
      _PFloat * pptr;
      _PFloat sqsum;
      _PFloat length;
      for(int i = 0; i < no_genes; ++i){
        pptr = &(*this)[i][0];
        sqsum = 0.0;
        for(int j = 0; j < vector_length; j++) sqsum += pptr[j] * pptr[j];
        length = sqrt(sqsum); 
        for(int j = 0; j < vector_length; j++) pptr[j] /= length;
      }
    }


    __INLINE Gene operator [] (int n){
      return Gene(this, n);
    }
    
    __INLINE iterator begin(){
      return iterator (Gene(this, 0));
    }

    __INLINE reverse_iterator rbegin(){
      return
          reverse_iterator (Gene(this, this->no_genes - 1));
    }

    __INLINE iterator end(){
      return iterator (Gene(this, this->no_genes));
    }

    __INLINE reverse_iterator rend(){
      return reverse_iterator (Gene(this, -1));
    }

  };

  __INLINE _PFloat & Gene::operator[] (int vector_no){
    return this->m->get(this->v, vector_no);
  }

  __INLINE const char * Gene::get_name(){
   return this->m->get_name(this->v);
  }
}


#endif /* MICROARRAY_HPP_*/
