#ifndef GRAH_HPP_
#define GRAH_HPP_

#include "config.h"
#include "constants.h"
#include "graph_interface.hpp"
#include <stdlib.h>
#include <string>
#include <assert.h>
#include <string.h>
#include <vector>
#include <map>
#include <cmath>
#include <limits>

using std::vector;
using std::map;
using std::string;
using std::numeric_limits;

namespace graph {

  template<class _P>
  class Graph : public allinone::Graph{
  
    private:
    _P * edge_matrix;
    size_t pitchx;
    size_t pitchy;
    long vector_length; 
    long no_vertices;

    virtual void init (long vector_length, long no_vertices){
      long dataSize = (long)this->no_vertices * this->no_vertices * 
                      this->vector_length * sizeof(_P);
      edge_matrix = (_P *)malloc(dataSize);
      pitchx = this->vector_length;
      pitchy = pitchx * this->no_vertices;
      //TODO this means that the graph is full
      memset(edge_matrix, 0, dataSize);
//      LOG(DEBUG, "Init");
    }

    public:
    Graph(bool is_directed, bool is_weighted, long vector_length, 
          long no_vertices, _PFloat threshold=numeric_limits<_P>::max()) : 
        allinone::Graph(is_directed, is_weighted, vector_length, 
          no_vertices, threshold, Graph_matrix, 
          allinone::Graph::get_datatype<_P>()),
        vector_length(vector_length), no_vertices(no_vertices) {
//      LOG(DEBUG, "Constructor");
      init(vector_length, no_vertices); 
    }

    Graph(allinone::Graph * _g, long no_vertices, bool copy_data = true) : 
        allinone::Graph(_g->get_is_directed(), _g->get_is_weighted(), 
        _g->get_vector_length(), no_vertices, _g->get_threshold(),Graph_matrix,
        allinone::Graph::get_datatype<_P>()), 
        vector_length(_g->get_vector_length()), no_vertices(no_vertices)
    { 
//      LOG(DEBUG, "Constructor");
//      *this = *_g;
      if(no_vertices < this->no_vertices){
        exit(-1);
      }
      init(vector_length, no_vertices);
      
      int iend = _g->get_no_vertices(); 
      this->set_names(&_g->get_name(0), iend);
      this->set_descs(&_g->get_desc(0), iend);
      
      if(copy_data)
        this->set_edges_from_graph(_g);
    }

    virtual ~Graph(){
      free (edge_matrix);
    }

    long get_pitchx(){return pitchx;}
    long get_pitchy(){return pitchy;}
    long get_vector_length() const {return vector_length;}
    long get_no_vertices()   const {return no_vertices;}
    
    inline _PFloat get(int from, int to, int vector_no) const{
      return *((Graph<_P> *) this)->get_t(from, to, vector_no);
    }

    inline _P * get_t(int from, int to, int vector_no){
#ifdef __DEBUG__
      if(min(from, to) < 0 || max(from, to) >= this->no_vertices || 
            vector_no < 0 || vector_no >= this->vector_length){ 
        LOG(WARNING, "Trying to access a vertice[%d][%d][%d]"
                   " which is not in the graph[%d][%d][%d]", 
                   from, to, vector_no, 
                   this->no_vertices, this->no_vertices, this->vector_length);
      }
#endif /*__DEBUG__*/
      return &edge_matrix[to * pitchx + from * pitchy + vector_no];
    } 

    template<class T>
    void set_t(int from, int to, int vector_no, T v){
#ifdef __DEBUG__
      if(min(from, to) < 0 || max(from, to) >= this->no_vertices || 
            vector_no < 0 || vector_no >= this->vector_length){ 
        LOG(DEBUG, "Trying to access a vertice[%d][%d][%d]"
                   " which is not in the graph[%d][%d][%d]", 
                   from, to, vector_no, 
                   this->no_vertices, this->no_vertices, this->vector_length);
      }
#endif /*__DEBUG__*/
      edge_matrix[to * pitchx + from * pitchy + vector_no] = v;
    }

    inline void set(int from, int to, int vector_no, _PFloat v){
      set_t(from, to, vector_no, v);
    }

    inline void set(int from, int to, int vector_no, int v){
      set_t(from, to, vector_no, v);
    }


//    void * graph(){
//      return this;
//    }

    inline bool is_edge(int from, int to, int vector_no) const{
#ifdef __DEBUG__
      if(min(from, to) < 0 || max(from, to) >= this->no_vertices || 
            vector_no < 0 || vector_no >= this->vector_length){ 
        LOG(WARNING, "Trying to access a vertex[%d][%d][%d]"
                   " which is not in the graph[%d][%d][%d]", 
                   from, to, vector_no, 
                   this->no_vertices, this->no_vertices, this->vector_length);
      }
#endif /*__DEBUG__*/
      return std::abs(this->get(from, to, vector_no))
              < this->get_threshold();
    }
   
  };
  
  allinone::Graph * allocate_graph(bool directed, bool weighted, 
        long vector_length, long no_vertices, string data_type, string type, 
        _PFloat threshold = numeric_limits<int>::max());
  
  allinone::Graph * allocate_graph(allinone::Graph * g, long no_vertices, 
      string data_type, string type, bool copy_data = true);

  //some functions
  int test();
  int test_time(int);
}

#endif /*GRAH_HPP_*/

