#include"timer/timer.h"
#include"graph_interface.hpp"
#include"results.hpp"
#include"multiple.hpp"
#include"constants.h"
#include<stdio.h>
#include<stdlib.h>
#include<assert.h>
#include<set>
#include<limits>
#include<sys/time.h>
#include<vector>
#include<string>
#include<algorithm>
#include<unistd.h>
#include<boost/lexical_cast.hpp>

using allinone::ArgAtoms;
using allinone::ArgAtom;
using allinone::Packages;
using allinone::Package;
using allinone::toInt;
using allinone::Graph;
using allinone::Results;
using allinone::Multiple;

typedef std::vector<int> assigned_t;
typedef std::vector<double> center_t;
typedef std::vector<center_t> centers_t;
typedef std::set<int> cluster_t;
typedef std::vector<cluster_t> clusters_t;
typedef std::vector<double> scores_t;

struct clusters_conf_t{
  assigned_t assigned;
  center_t center;
  centers_t centers;
  cluster_t cluster;
  clusters_t clusters;
  scores_t scores;
};
    
namespace kmeans{
  void kmeans_assign_seeds(Graph *, clusters_t *, assigned_t *, string);
}

namespace graph{
  void save_mat(FILE * f, string name, allinone::Graph * g, char delimit);
}

namespace kmeans_euc{

  static int seed_initialized = 0;

  void kmeans_calculate_center(Graph * g, clusters_conf_t * cc, int cluster){
    int no_verts = g->get_no_vertices();
    int no_el = cc->clusters.at(cluster).size();
    center_t & center = cc->centers.at(cluster);
    foreach(double & val, center) val = 0;

    foreach(int val, cc->clusters.at(cluster)) 
    for(int i = 0; i < no_verts; ++i)
      center.at(i) += g->get(val, i, 0); 
      
    foreach(double & val, center) val /= no_el;
  }

  void kmeans_calculate_centers(Graph * g, clusters_conf_t * cc){
    for(int i = 0, iend = cc->clusters.size(); i < iend; ++i)  
      kmeans_calculate_center(g, cc, i);
  }

  void kmeans_recalculate_sumdistance(Graph * g, clusters_conf_t * cc, int cluster, double * score){
    double ret = 0;
    center_t & c = cc->centers.at(cluster);
    for(int i =0, iend = cc->clusters.size(); i < iend; ++i){
      if(i != cluster){
        double distance = 0;
        center_t &c1 = cc->centers.at(i);
        for(int j = 0, jend = c.size(); j < jend; ++j){
          double tmp = c.at(j) - c1.at(j);
          distance+= tmp * tmp;
        }
        ret += sqrt(distance);
      }
    }
    *score = ret / (cc->clusters.size() - 1);
  }
  
  void kmeans_recalculate_sumdistances(Graph * g, clusters_conf_t * cc ){
    for(int i = 0, iend = cc->clusters.size(); i < iend; ++i)
      kmeans_recalculate_sumdistance(g, cc, i, &cc->scores.at(i));
  }

  void kmeans_calculate_distance(Graph * g, clusters_conf_t * cc, int cluster, int from, double * dist){
    double distance = 0;
    center_t & c = cc->centers.at(cluster);
    for(int j = 0, jend = c.size(); j < jend; ++j){
      double tmp = c.at(j) - g->get(from, j, 0);
      distance+= tmp * tmp;
    }
    *dist = sqrt(distance);
  }

  int kmeans_choose_best(Graph * g, clusters_conf_t * cc, int from){
    scores_t distances (cc->clusters.size(), std::numeric_limits<double>::max());

    for(int i = 0, iend = cc->clusters.size(); i < iend; ++i)
      kmeans_calculate_distance(g, cc, i, from, &distances.at(i));
    
    double best_val = std::numeric_limits<double>::max(); 
    int best_i = 0; 
    for(int i = 0, iend = distances.size(); i < iend; ++i)
      if(distances.at(i) < best_val) best_val = distances.at(best_i = i);

    return best_i;
  }

  void kmeans_add_to_cluster(Graph * g, clusters_conf_t * cc, int v, int cluster){
    cluster_t & cl = cc->clusters[cluster];
    center_t  & cn = cc->centers[cluster];
    int no_el = cl.size();
    int no_verts = g->get_no_vertices();
    for(int i = 0; i < no_verts; ++i)
      cn.at(i) = (cn.at(i) * no_el + g->get(v, i, 0)) / (no_el + 1); 
    cl.insert(v);
    cc->assigned[v] = cluster;
  }

  void kmeans_assign_rest(Graph * g, clusters_conf_t * cc){
    for(int i = 0, iend = cc->assigned.size(); i < iend; ++i){
      int to = cc->assigned.at(i);
      if(to == -1)
        kmeans_add_to_cluster(g, cc, i, kmeans_choose_best(g, cc, i));
    }
  }

  Results * kmeans_lloyd(Graph * g, int n_of_clusters, int iteration, string in_assignment) {
    int nverts = g->get_no_vertices();
    assert(n_of_clusters <= nverts);

    clusters_conf_t cc;

    cc.assigned.resize(nverts, -1);
    cc.clusters.resize(n_of_clusters);

    cc.centers.resize(n_of_clusters);
    foreach(center_t & center, cc.centers)
      center.resize(nverts);

    //these are actually distances. The score is assumed to be the higher, the better.
    cc.scores.resize(n_of_clusters, 0);
    
    LOG(DEBUG, "Beginning finding optimal ;) %d clusters, using the kmeans lloyd method based on vectors", n_of_clusters);
    kmeans::kmeans_assign_seeds(g, &cc.clusters, &cc.assigned, in_assignment);
    if(verbosity >= DEBUG){
      std::string seeds; seeds.reserve(1000);
      for(int i = 0; i < n_of_clusters; ++i) 
        seeds += " " +
          boost::lexical_cast<std::string>(*cc.clusters[i].begin());
      LOG(DEBUG, "Chosen seeds: %s", seeds.c_str());
    }

    kmeans_calculate_centers(g, &cc);
    kmeans_assign_rest(g, &cc);
    
    double sum2, sum;
    kmeans_recalculate_sumdistances(g, &cc);
    sum = 0; foreach(double val, cc.scores) sum += val;

    if(verbosity >= DEBUG){
      std::string clusts; clusts.reserve(1000);
      for(int i = 0; i < n_of_clusters; ++i) 
        clusts += " " + boost::lexical_cast<std::string>(cc.clusters[i].size());
      LOG(DEBUG, "Initial clusters' distances: %g and sizes: %s", sum, 
                                                    clusts.c_str());
    }

    int it = 0;
    for(int i = 0; i!= iteration; ++i){
      sum2 = sum;

      for(int i = 0; i < nverts; ++i){
        int & to = cc.assigned.at(i);
        int to2 = kmeans_choose_best(g, &cc, i);
        if(to != to2){
          cc.clusters[to].erase(i);
          cc.clusters[to = to2].insert(i);
        }
      }

      kmeans_calculate_centers(g, &cc);
      kmeans_recalculate_sumdistances(g, &cc);
      sum = 0; foreach(double val, cc.scores) sum += val;

      if(verbosity >= DEBUG){
        std::string clusts; clusts.reserve(1000);
        for(int i = 0; i < n_of_clusters; ++i) 
          clusts += " " + boost::lexical_cast<std::string>(cc.clusters[i].size());
        LOG(DEBUG, "It(%d) clusters' distances: %g and sizes: %s", it++, sum, 
                                                      clusts.c_str());
      }
      if(sum2 == sum) break;
    }
    
    Results * res = new Results(sum, g->get_no_vertices()); 
    res->set_names(&g->get_name(0), g->get_no_vertices(), 0);
    res->set_descs(&g->get_desc(0), g->get_no_vertices(), 0);
    deque<Results::Result> & deqres = res->get_deq();
    
    for(int i = 0, iend = cc.clusters.size(); i < iend; ++i){
      deqres.push_back(Results::Result());
      Results::Result & r = deqres.back();
      r.first = cc.scores[i];
      foreach(int j, cc.clusters[i]) r.second[j] = 0;
    }

    return res;
  }
  
  Results * kmeans_r(Graph * g, int n_of_clusters, int iteration, int n_of_tries, string in_assignment) {
    char name1[100] = "tmp_infile_XXXXXX";
    char name2[100] = "tmp_outfile_XXXXXX";
    int fd1 = mkstemp(name1), fd2 = mkstemp(name2);
    assert(fd1 != -1 && fd2 !=-1);
    close(fd2);

    FILE * f_mat = fdopen(fd1, "w"); assert(f_mat);
    graph::save_mat(f_mat, name1, g, ' ');
    fclose(f_mat);
          
    char command[1000] = {0};
    int n = snprintf(command, sizeof(command),
        "echo 't=read.table(\"%s\")\nwrite(kmeans(t, %d, %d, %d)$cluster, \"%s\")' | R --no-save >/dev/null",
        name1, n_of_clusters, iteration == -1 ? 10 : iteration, n_of_tries, name2);
    assert(n < sizeof(command));
    LOG(INFO, "Executing command:\n%s", command);
    int ret = system(command);
    assert(!ret);
    Results * res = kmeans_lloyd(g, n_of_clusters, 0, name2);
    assert(!unlink(name1)); assert(!unlink(name2));
    return res;
  }

}



